@@ -2,12 +2,12 @@
 from collections import OrderedDict, defaultdict
 from functools import partial
 from itertools import chain
-from typing import TYPE_CHECKING, Any, Dict, List, NamedTuple, Optional, Tuple
+from typing import TYPE_CHECKING, Any, Dict, List, NamedTuple, Optional, Set, Tuple
 from warnings import warn
 
 from . import place
 from .comments import parse as parse_comments
-from .deprecated.finders import FindersManager
+from .exceptions import MissingSection
 from .settings import DEFAULT_CONFIG, Config
 
 if TYPE_CHECKING:
@@ -31,10 +31,9 @@ if TYPE_CHECKING:
 def _infer_line_separator(contents: str) -> str:
     if "\r\n" in contents:
         return "\r\n"
-    elif "\r" in contents:
+    if "\r" in contents:
         return "\r"
-    else:
-        return "\n"
+    return "\n"
 
 
 def _normalize_line(raw_line: str) -> Tuple[str, str]:
@@ -55,11 +54,11 @@ def import_type(line: str, config: Config = DEFAULT_CONFIG) -> Optional[str]:
     """If the current line is an import line it will return its type (from or straight)"""
     if config.honor_noqa and line.lower().rstrip().endswith("noqa"):
         return None
-    elif "isort:skip" in line or "isort: skip" in line or "isort: split" in line:
+    if "isort:skip" in line or "isort: skip" in line or "isort: split" in line:
         return None
-    elif line.startswith(("import ", "cimport ")):
+    if line.startswith(("import ", "cimport ")):
         return "straight"
-    elif line.startswith("from "):
+    if line.startswith("from "):
         return "from"
     return None
 
@@ -139,6 +138,7 @@ class ParsedContent(NamedTuple):
     line_separator: str
     sections: Any
     verbose_output: List[str]
+    trailing_commas: Set[str]
 
 
 def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedContent:
@@ -151,6 +151,8 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
     out_lines = []
     original_line_count = len(in_lines)
     if config.old_finders:
+        from .deprecated.finders import FindersManager
+
         finder = FindersManager(config=config).find
     else:
         finder = partial(place.module, config=config)
@@ -175,6 +177,8 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
         "above": {"straight": {}, "from": {}},
     }
 
+    trailing_commas: Set[str] = set()
+
     index = 0
     import_index = -1
     in_quote = ""
@@ -186,8 +190,10 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
             line, in_quote=in_quote, index=index, section_comments=config.section_comments
         )
 
-        if line in config.section_comments and not skipping_line:
-            if import_index == -1:
+        if (
+            line in config.section_comments or line in config.section_comments_end
+        ) and not skipping_line:
+            if import_index == -1:  # pragma: no branch
                 import_index = index - 1
             continue
 
@@ -235,7 +241,6 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
                             and not commentless.rstrip().endswith(")")
                             and import_index < line_count
                         ):
-
                             while import_index < line_count and not commentless.rstrip().endswith(
                                 ")"
                             ):
@@ -246,8 +251,8 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
 
                         if import_index >= line_count:
                             break
-                        else:
-                            starting_line = in_lines[import_index]
+
+                        starting_line = in_lines[import_index]
 
         line, *end_of_line_comment = line.split("#", 1)
         if ";" in line:
@@ -260,6 +265,7 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
         for statement in statements:
             line, raw_line = _normalize_line(statement)
             type_of_import = import_type(line, config) or ""
+            raw_lines = [raw_line]
             if not type_of_import:
                 out_lines.append(raw_line)
                 continue
@@ -288,9 +294,11 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
                     ):
                         nested_comments[stripped_line] = comments[-1]
                     import_string += line_separator + line
+                    raw_lines.append(line)
             else:
                 while line.strip().endswith("\\"):
                     line, new_comment = parse_comments(in_lines[index])
+                    line = line.lstrip()
                     index += 1
                     if new_comment:
                         comments.append(new_comment)
@@ -310,6 +318,7 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
                         ):
                             nested_comments[stripped_line] = comments[-1]
                         import_string += line_separator + line
+                        raw_lines.append(line)
 
                         while not line.split("#")[0].strip().endswith(")") and index < line_count:
                             line, new_comment = parse_comments(in_lines[index])
@@ -325,6 +334,7 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
                             ):
                                 nested_comments[stripped_line] = comments[-1]
                             import_string += line_separator + line
+                            raw_lines.append(line)
 
                     stripped_line = _strip_syntax(line).strip()
                     if (
@@ -348,6 +358,10 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
                     .replace("\\", " ")
                     .replace("\n", " ")
                 )
+                if "import " not in import_string:
+                    out_lines.extend(raw_lines)
+                    continue
+
                 if " cimport " in import_string:
                     parts = import_string.split(" cimport ")
                     cimports = True
@@ -369,6 +383,7 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
             attach_comments_to: Optional[List[Any]] = None
             direct_imports = just_imports[1:]
             straight_import = True
+            top_level_module = ""
             if "as" in just_imports and (just_imports.index("as") + 1) < len(just_imports):
                 straight_import = False
                 while "as" in just_imports:
@@ -384,7 +399,7 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
                         direct_imports.remove("as")
                         if nested_module == as_name and config.remove_redundant_aliases:
                             pass
-                        elif as_name not in as_map["from"][module]:
+                        elif as_name not in as_map["from"][module]:  # pragma: no branch
                             as_map["from"][module].append(as_name)
 
                         full_name = f"{nested_module} as {as_name}"
@@ -393,7 +408,7 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
                             categorized_comments["nested"].setdefault(top_level_module, {})[
                                 full_name
                             ] = associated_comment
-                            if associated_comment in comments:
+                            if associated_comment in comments:  # pragma: no branch
                                 comments.pop(comments.index(associated_comment))
                     else:
                         module = just_imports[as_index - 1]
@@ -409,9 +424,16 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
                                 f"{top_level_module}.__combined_as__", []
                             )
                         else:
-                            attach_comments_to = categorized_comments["straight"].setdefault(
-                                module, []
-                            )
+                            if type_of_import == "from" or (
+                                config.remove_redundant_aliases and as_name == module.split(".")[-1]
+                            ):
+                                attach_comments_to = categorized_comments["straight"].setdefault(
+                                    module, []
+                                )
+                            else:
+                                attach_comments_to = categorized_comments["straight"].setdefault(
+                                    f"{module} as {as_name}", []
+                                )
                     del just_imports[as_index : as_index + 2]
 
             if type_of_import == "from":
@@ -429,6 +451,10 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
                         f"could not place module {import_from} of line {line} --"
                         " Do you need to define a default section?"
                     )
+
+                if placed_module and placed_module not in imports:
+                    raise MissingSection(import_module=import_from, section=placed_module)
+
                 root = imports[placed_module][type_of_import]  # type: ignore
                 for import_name in just_imports:
                     associated_comment = nested_comments.get(import_name)
@@ -436,14 +462,28 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
                         categorized_comments["nested"].setdefault(import_from, {})[
                             import_name
                         ] = associated_comment
-                        if associated_comment in comments:
+                        if associated_comment in comments:  # pragma: no branch
                             comments.pop(comments.index(associated_comment))
+                if (
+                    config.force_single_line
+                    and comments
+                    and attach_comments_to is None
+                    and len(just_imports) == 1
+                ):
+                    nested_from_comments = categorized_comments["nested"].setdefault(
+                        import_from, {}
+                    )
+                    existing_comment = nested_from_comments.get(just_imports[0], "")
+                    nested_from_comments[
+                        just_imports[0]
+                    ] = f"{existing_comment}{'; ' if existing_comment else ''}{'; '.join(comments)}"
+                    comments = []
 
                 if comments and attach_comments_to is None:
                     attach_comments_to = categorized_comments["from"].setdefault(import_from, [])
 
                 if len(out_lines) > max(import_index, 1) - 1:
-                    last = out_lines and out_lines[-1].rstrip() or ""
+                    last = out_lines[-1].rstrip() if out_lines else ""
                     while (
                         last.startswith("#")
                         and not last.endswith('"""')
@@ -477,6 +517,9 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
 
                 if comments and attach_comments_to is not None:
                     attach_comments_to.extend(comments)
+
+                if "," in import_string.split(just_imports[-1])[-1]:
+                    trailing_commas.add(import_from)
             else:
                 if comments and attach_comments_to is not None:
                     attach_comments_to.extend(comments)
@@ -488,8 +531,7 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
                         comments = []
 
                     if len(out_lines) > max(import_index, +1, 1) - 1:
-
-                        last = out_lines and out_lines[-1].rstrip() or ""
+                        last = out_lines[-1].rstrip() if out_lines else ""
                         while (
                             last.startswith("#")
                             and not last.endswith('"""')
@@ -524,6 +566,10 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
                             " Do you need to define a default section?"
                         )
                         imports.setdefault("", {"straight": OrderedDict(), "from": OrderedDict()})
+
+                    if placed_module and placed_module not in imports:
+                        raise MissingSection(import_module=module, section=placed_module)
+
                     straight_import |= imports[placed_module][type_of_import].get(  # type: ignore
                         module, False
                     )
@@ -545,4 +591,5 @@ def file_contents(contents: str, config: Config = DEFAULT_CONFIG) -> ParsedConte
         line_separator=line_separator,
         sections=config.sections,
         verbose_output=verbose_output,
+        trailing_commas=trailing_commas,
     )
