@@ -85,7 +85,7 @@ def test_colored_printer_diff(capsys):
     assert colorama.Fore.GREEN + "+ added line" in out
     # Removed lines are red
     assert colorama.Fore.RED + "- removed line" in out
-    # Normal lines are resetted back
+    # Normal lines are reset back
     assert colorama.Style.RESET_ALL + "normal line" in out
 
 
@@ -105,7 +105,7 @@ def test_colored_printer_diff_output(capsys):
 def test_colorama_not_available_handled_gracefully(capsys):
     with pytest.raises(SystemExit) as system_exit:
         _ = isort.format.create_terminal_printer(color=True)
-    assert system_exit.value.code > 0
+    assert system_exit.value.code and int(system_exit.value.code) > 0
     _, err = capsys.readouterr()
     assert "colorama" in err
     assert "colors extra" in err
